from __future__ import print_function

import logging
import leip
import os
import re

from mad2.util import get_all_mad_files

lg = logging.getLogger(__name__)


@leip.hook("madfile_post_save", 100)
def root(app, madfile):
    """
    Set owner & permission (if allowed to do so)

    """

    filename = madfile['inputfile']
    madname = madfile.get('madname')

    #change SHA1SUM file
    dirname = madfile['dirname']
    sha1name = os.path.join(dirname, 'SHA1SUMS')
    metaname = os.path.join(dirname, 'SHA1SUMS.META')

    if not os.path.exists(dirname):
        #can't do much if there is not containing directory
        return

    dstats = os.stat(dirname)


    #attempt to set  permissions as non root
    try:
        if not madname is None and \
                os.path.exists(madname):
            fstats = os.stat(filename)
            os.chmod(madname, fstats.st_mode)
    except:
        pass

    try:
        if os.path.exists(sha1name):
            os.chmod(sha1name, 0o664)
        if os.path.exists(metaname):
            os.chmod(metaname, 0o664)
    except:
        pass

    if os.geteuid() != 0:
        #only do something when root
        return


    if not madname is None and \
            os.path.exists(madname):
        fstats = os.stat(filename)
        os.chown(madname, fstats.st_uid, fstats.st_gid)

    if os.path.exists(sha1name):
        os.chown(sha1name, dstats.st_uid, dstats.st_gid)

    if os.path.exists(metaname):
        os.chown(metaname, dstats.st_uid, dstats.st_gid)




import errno
import logging
import os
import re

import leip

from mad2 import ui

lg = logging.getLogger(__name__)

def isint(n):
    try:
        int(n)
        return True
    except ValueError:
        return False

def dehumanize(n):
    n = n.strip().upper()
    ff = re.match("([0-9]+)([KMGT]?)", n)
    if not ff:
        lg.warning("Invalid size specification: %s", n)
        exit(-1)
    base, ext = ff.groups()
    if not ext:
        return int(base)
    mult = 1024 ** (1 + 'KMGT'.index(ext))
    return int(float(base) * mult)


@leip.arg('-I', '--ignore_mad_ignore', action='store_true',
          help='ignore mad.ignore files')
@leip.arg('-s', '--minsize', help='minimum size (may use 1k, 1M, 1G)' +
          ' default: 10k', default='10k')
@leip.arg('-n', '--no_recurse', action='store_true',
          help='no recursive search')
@leip.arg('-p', '--progress', action='store_true',
          help='show some progress indicator')
@leip.arg('-d', '--do_dot_dirs', action='store_true',
          help='do not ignore .* directories')
@leip.arg('-a', '--do_dot_files', action='store_true',
          help='do not ignore .* files')
@leip.command
def scan(app, args):
    """
    (recursively) find and print files may have a mad annotation
    """
    minsize = dehumanize(args.minsize)

    files_to_ignore = app.conf.get('plugin.find.ignore', [])

    def check_write_permission(d):
        sha1file = os.path.join(d, 'SHA1SUMS')
        # we'll assume that if we have access to the sha1 file we also
        # have access to the meta file
        if os.path.exists(sha1file):
            if os.access(sha1file, os.W_OK):
                return True
            else:
                lg.debug("No write permission on %s", sha1file)
                return False

        #no sha1file - then check for write access on the directory
        return os.access(d, os.W_OK)

    app.trans['progress.find'] = 0
    counter = 0
    #lg.setLevel(logging.DEBUG)
    for dirpath, dirnames, filenames in os.walk('.'):

        #ui.message("considering %s (%d dirs, %d files)",
        #           os.path.basename(dirpath), len(dirnames),
        #           len(filenames))

        # if it's unlikely that we're able to write sha1sums to a
        # local file, we're not going to process this file
        if not check_write_permission(dirpath):
            continue

        dirs_to_remove = set()

        #never traverse into a .mad directorie
        while '.mad' in dirnames:
            dirnames.remove('.mad')

        #remove all .* directories
        for d in dirnames:
            if (not args.do_dot_dirs) and d[0] == '.':
                dirs_to_remove.add(d)
            if not args.ignore_mad_ignore:
                madignore = os.path.join(
                    dirpath, d, 'mad.ignore')
                if os.path.exists(madignore):
                    dirs_to_remove.add(d)

        for d in dirs_to_remove:
            dirnames.remove(d)

        lg.debug(" - have %d subdirs left", len(dirnames))
        lg.debug(" - processing %d files", len(filenames))

        for f in filenames:

            ffn = os.path.join(dirpath, f)
            #lg.debug("considering file: %s", f)

            if f in files_to_ignore:
                continue

            if (not args.do_dot_files) and \
                    f[0] == '.':
                continue

            try:
                filestat = os.stat(ffn)
            except OSError, e:
                lg.debug(" --- cannot stat: %s (%d)", f, e.errno)
                if e.errno == errno.ENOENT:
                    #path does not exists - or is a broken symlink
                    continue
                else:
                    lg.warning("error scanning %s", ffn)
                    lg.warning("%s", e)
                    continue

            if minsize > 0:
                if filestat.st_size < minsize:
                    #lg.debug(" --- too small: %s (%d)", f, filestat.st_size)
                    continue

            if not os.access(ffn, os.R_OK):
                #no read permission
                continue

            app.trans['progress.find'] += 1
            counter += 1

            pp = app.trans['progress.find']
            if args.progress and  pp > 0 and pp % 2500 == 0:
                lg.warning("mad scan: found {} files".format(pp))
                lg.warning("  now in: {}".format(dirpath))

            print(ffn)

        if dirpath == '.' and args.no_recurse:
            break




import fnmatch
import logging
import os

import leip

import mad2.util

lg = logging.getLogger(__name__)
lg.setLevel(logging.DEBUG)


@leip.subparser
def inotify(app, args):
    """ Inotify fs monitoring
    """
    pass


@leip.arg('-s', '--minsize', help='minimum size for a file to be considered', 
          type=int, default=100)
@leip.arg('path')
@leip.subcommand(inotify, 'watch')
def inotify_watch(app, args):
    
    import pyinotify

    parent_paths_to_ignore = []

    def must_be_saved(path):

        if os.path.isdir(path):
            return False

        base = os.path.basename(path)
        for pattern in app.conf['ignore.filename']:
            if fnmatch.fnmatch(base, pattern):
                #lg.debug("ignoring (base) %s", path)
                return False

        thisdir = os.path.dirname(path).rstrip('/')

        while True:
            thisbase = os.path.basename(thisdir)
            if thisdir in parent_paths_to_ignore:
                #lg.debug("Ignoring path cache: %s", path)
                return False

            if thisbase in  app.conf['ignore.inpath']:
                parent_paths_to_ignore.append(thisdir)
                return False

            for pp in app.conf['ignore.parentpattern']:
                pptest = os.path.join(thisdir, pp)
                if os.path.exists(pptest):
                    #lg.debug("Ignoring: %s because of %s", path, pptest)
                    parent_paths_to_ignore.append(thisdir)
                    return False

            thisdir = os.path.dirname(thisdir).rstrip('/')
            if not thisdir:
                break
                                  
        return True

        

    class MadEventHandler(pyinotify.ProcessEvent):

        def process_IN_DELETE(self, event):
            lg.info("Delete: %s", event.pathname)
            maf = mad2.util.get_mad_file(app, event.pathname)
            maf.delete()
            maf.flush()

        def process_IN_MOVED_FROM(self, event):
            return self.process_IN_DELETE(event)

        def process_default(self, event):

            pn = event.pathname

            try:
                if os.path.exists(pn) and \
                   os.path.getsize(pn) < args.minsize:
                    return
                
                if not must_be_saved(pn):
                    return 

            except Exception as e:
                lg.error("problem processing %s", pn)
                lg.error(" - %s", str(e))
                return

            try:
                lg.debug("saving: %s", event.pathname)
                maf = mad2.util.get_mad_file(app, event.pathname)
                maf.save()
                maf.flush()
            except Exception as e:
                lg.error("problem mad saving %s", pn)
                lg.error(" - %s", str(e))
                

    mask = pyinotify.IN_DELETE \
        | pyinotify.IN_MOVED_FROM \
        | pyinotify.IN_CREATE \
        | pyinotify.IN_MODIFY \
        | pyinotify.IN_ATTRIB \
        | pyinotify.IN_MOVED_TO 


    wm = pyinotify.WatchManager()

    handler = MadEventHandler()
    notifier = pyinotify.Notifier(wm, handler)
    wm.add_watch(args.path, mask, rec=True)
    notifier.loop()






from collections import defaultdict

import leip

from mad2 import util

from IPython.core.magic import magics_class, line_magic, cell_magic, Magics


@magics_class
class MadMadMad(Magics):

    def __init__(self, *args, **kwargs):
        super(MadMadMad, self).__init__(*args, **kwargs)

        self.madapp = leip.app('mad2')
        
        self.input = set()
        self.output = set()
        self.db = set()


    @line_magic
    def mad_provenance_input(self, line):
        "my line magic"
        files = line.strip().split()
        for f in line.strip().split():
            self.input.add(f)

    @line_magic
    def mad_provenance_output(self, line):
        "my line magic"
        files = line.strip().split()
        for f in line.strip().split():
            self.output.add(f)

    @line_magic
    def mad_provenance_db(self, line):
        "my line magic"
        files = line.strip().split()
        for f in line.strip().split():
            self.db.add(f)
        
    @line_magic
    def mad_provenance_register(self, line):
        mf = defaultdict(list)
        
        for fn in self.input:
            mf['input'].append(util.get_mad_file(self.madapp, fn))
        for fn in self.output:
            mf['output'].append(util.get_mad_file(self.madapp, fn))
        for fn in self.db:
            mf['db'].append(util.get_mad_file(self.madapp, fn))

        rv = []
        import os
        rv.append(os.getcwd())
        ip = get_ipython()
        rv.append(dir(ip))
        rv.append(ip)
        for cat in mf:
            for madfile in mf[cat]:
                madfile.load()

#                rv.append(cat)
#                rv.append(madfile)
#                rv.append(madfile['orphan'])
#                rv.append(madfile['fullpath'])
                #1madfile.save()
            
        return map(str, rv)
                      
def load_ipython_extension(ip):
    ip.register_magics(MadMadMad)
            





import collections
import functools

import cPickle
import errno
import logging
import os
import re
import sys
import time

from termcolor import cprint

from mad2.exception import MadPermissionDenied, MadNotAFile
from mad2.madfile import MadFile, MadDummy

import fantail


lg = logging.getLogger(__name__)
#lg.setLevel(logging.DEBUG)

#
# Helper function - instantiate a madfile, and provide it with a
# method to run hooks
#

STORES = None



def persistent_cache(path, cache_on, duration):
    """
    Disk persistent cache that reruns a function once every
    'duration' no of seconds
    """
    def decorator(original_func):

        def new_func(*args, **kwargs):

            if isinstance(cache_on, str):
                cache_name = kwargs[cache_on]
            elif isinstance(cache_on, int):
                cache_name = args[cache_on]

            full_cache_name = os.path.join(path, cache_name)
            lg.debug("cache file: %s", full_cache_name)
            run = False

            if kwargs.get('force'):
                run = True

            if not os.path.exists(full_cache_name):
                #file does not exist. Run!
                run = True
            else:
                #file exists - but is it more recent than
                #duration (in seconds)
                mtime = os.path.getmtime(full_cache_name)
                age = time.time() - mtime
                if age > duration:
                    lg.debug("Cache file is too recent")
                    lg.debug("age: %d", age)
                    lg.debug("cache refresh: %d", duration)
                    run = True

            if not run:
                #load from cache
                lg.debug("loading from cache: %s", full_cache_name)
                with open(full_cache_name) as F:
                    res = cPickle.load(F)
                    return res


            #no cache - create
            lg.debug("no cache - running function %s", original_func)
            rv = original_func(*args, **kwargs)
            lg.debug('write to cache: %s', full_cache_name)

            if not os.path.exists(path):
                os.makedirs(path)
            with open(full_cache_name, 'wb') as F:
                cPickle.dump(rv, F)

            return rv

        return new_func

    return decorator


def initialize_stores(app):
    # prevent circular import
    import mad2.store

    global STORES
    STORES = {}
    for store in app.conf['store']:
        store_conf = app.conf['store'][store]
        if not isinstance(store_conf, dict):
            continue
        if not store_conf.get('enabled', False):
            continue
        STORES[store] = mad2.store.all_stores[store](store_conf)


def cleanup_stores(app):
    if STORES is None:
        return

    for store_name in STORES:
        store = STORES[store_name]
        store.finish()


def get_mad_file(app, filename):
    """
    Instantiate a mad file & add hooks
    """
    global STORES
    if STORES is None:
        initialize_stores(app)

    lg.debug("instantiating madfile for {0}".format(filename))
    return MadFile(filename,
                   stores=STORES,
                   base=app.conf['madfile'],
                   hook_method=app.run_hook)


def get_mad_dummy(app, data):
    """
    instantiate a dummy - only used to save.

    """

    global STORES
    if STORES is None:
        initialize_stores(app)

    if 'madname' in data:
        del data['madname']
    if 'fullmadpath' in data:
        del data['fullmadpath']

    data_all = fantail.Fantail(data)
    data_core = fantail.Fantail()

    for kw in data_all.keys():
        tra = app.conf['keywords'][kw].get('transient', False)
        if not tra:
            data_core[kw] = data_all[kw]

    lg.debug("instantiating dummy madfile")
    return MadDummy(data_all=data_all, data_core=data_core, stores=STORES,
                    hook_method=app.run_hook )

def to_mad(fn):
    if '/' in fn:
        a, b = fn.rsplit('/')
        return os.path.join(a, '.{}.mad'.format(b))
    else:
        return '.{}.mad'.format(fn)


def get_filenames(args, use_stdin=True, allow_dirs=False):
    """
    Get all incoming filenames
    """
    filenames = []
    demad = re.compile(r'^(?P<path>.*/)?\.(?P<fn>[^/].+)\.mad$')

    def demadder(m):
        if not m.group('path') is None:
            return '{}{}'.format(m.group('path'), m.group('fn'))
        else:
            return m.group('fn')

    if 'file' in args and len(args.file) > 0:

        for f in args.file:
            if len(f) == 0:
                continue
            if '.mad/' in f:
                continue
            if 'SHA1SUMS' in f:
                continue
            if 'SHA1SUMS.META' in f:
                continue
            if 'QDSUMS' in f:
                continue


            if not os.access(f, os.R_OK):
                # no read access - ignore this file
                continue


            try:
                os.stat(f)
            except OSError, e:
                lg.warning("Problem getting stats from %s", f)
                if e.errno == errno.ENOENT:
                    # path does not exists - or is a broken symlink
                    continue
                else:
                    raise

            lg.debug("processing %s", f)

            rv = demad.sub(demadder, f)

            if not allow_dirs and os.path.isdir(rv):
                continue

            yield rv

    elif use_stdin:

        # nothing in args - see if there is something on stdin
        for line in sys.stdin:
            line = line.strip()
            if '.mad/' in line:
                continue
            rv = demad.sub(demadder, line)
            if os.path.isdir(rv):
                continue
            yield rv


def get_all_mad_files(app, args, use_stdin=True, warn_on_errors=True):
    """
    get input files from sys.stdin and args.file
    """

    for filename in get_filenames(args, use_stdin):
        try:
            maf = get_mad_file(app, filename)
            yield maf
        except MadNotAFile:
            pass
        except MadPermissionDenied:
            lg.warning("Permission denied: {}".format(
                filename))
        except Exception, e:
            if warn_on_errors:
                lg.warning("Error instantiating %s", filename)
                lg.warning("Error: %s", str(e))
                raise
            else:
                raise

# Thanks: http://tinyurl.com/kq5hxtr


def humansize(nbytes):
    import numpy as np
    if np.isnan(nbytes):
        return float("nan")
    suffixes = [' b', 'kb', 'Mb', 'Gb', 'Tt', 'Pb']
    if nbytes == 0:
        return '0 b'
    i = 0
    while nbytes >= 1024 and i < len(suffixes) - 1:
        nbytes /= 1024.
        i += 1
    f = ('%.2f' % nbytes).rstrip('.')
    return '%s %s' % (f, suffixes[i])


def boolify(v):
    """
    return a boolean from a string
    yes, y, true, True, t, 1 -> True
    otherwise -> False
    """
    return v.lower() in ['yes', 'y', 'true', 't', '1']


def message(cat, message, *args):
    if len(args) > 0:
        message = message.format(*args)

    message = " ".join(message.split())
    color = {'er': 'red',
             'wa': 'yellow',
             'in': 'green',
             }.get(cat.lower()[:2], 'blue')

    cprint('Kea', 'cyan', end="/")
    cprint(cat, color)
    for line in textwrap.wrap(message):
        print "  " + line


def render(txt, data):

    env = moa.moajinja.getStrictEnv()
    renconf = self.render()
    templ = env.from_string(value)
    try:
        rv = templ.render(renconf)
        return rv
    except jinja2.exceptions.UndefinedError:
        return value
    except jinja2.exceptions.TemplateSyntaxError:
        return value


# Borrowed from: http://tinyurl.com/majcr53
class memoized(object):

    '''Decorator. Caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned
    (not reevaluated).
    '''

    def __init__(self, func):
        self.func = func
        self.cache = {}

    def __call__(self, *args):
        if not isinstance(args, collections.Hashable):
            # uncacheable. a list, for instance.
            # better to not cache than blow up.
            return self.func(*args)
        if args in self.cache:
            return self.cache[args]
        else:
            value = self.func(*args)
            self.cache[args] = value
            return value

    def __repr__(self):
        '''Return the function's docstring.'''
        return self.func.__doc__

    def __get__(self, obj, objtype):
        '''Support instance methods.'''
        return functools.partial(self.__call__, obj)




from __future__ import print_function

import collections
import logging
from multiprocessing.dummy import Pool as ThreadPool
import os
import subprocess as sp
import uuid

import leip
import fantail

from mad2.util import get_all_mad_files
from mad2.plugin.template import render_numeric

lg = logging.getLogger(__name__)
# lg.setLevel(logging.DEBUG)


class Executor(object):

    def __init__(self, app, args):
        self.app = app
        self.args = args

        executor_type = args.executor

        self.executor_type = executor_type
        confName = executor_type.lower()

        self.conf = app.conf['x.executor.{}'.format(confName)]
        self.defaults = app.conf['x.executor.default']

        lg.debug("opening thread pool with %d threads",
                 args.threads)

        self.pool = ThreadPool(args.threads)

    def get_script_template(self, xtype):
        if xtype in self.conf:
            return self.conf[xtype]

        return self.defaults[xtype]

    def prepare_commandline(self, madfile, command_info):
        pass

    def prepare_script(self, madfileset, command_info):
        """
        prepare command line & full execution prepare_script

        """
        xtra_info = fantail.Fantail()
        xtra_info['input_file'] = madfileset[0]
        xtra_info['input_files'] = madfileset
        xtra_info['filename'] = madfileset[0]["filename"]
        xtra_info['comm'] = command_info

        lg.debug("filename: " + xtra_info['filename'])

        xtype = command_info.get('type', 'map')
        lg.debug("exec type %s", xtype)

        # first step - render the actual command
        x_conf = [xtra_info,
                  command_info['defaults'],
                  self.conf,
                  self.defaults,
                  self.app.conf['plugin.x.defaults'],
                  self.app.conf]

        # hack - should be possible from within jinja
        command = command_info['command']

        cl = madfileset[0].render(command, [madfileset[0]] + x_conf)

        xtra_info['cl'] = cl
        xtra_info['input_files'] = madfileset
        xtra_info['pwd'] = os.getcwd()
        xtra_info['comm'] = command_info
        xtra_info['uuid'] = str(uuid.uuid1())

        script = self.get_script_template(xtype)
        script = madfileset[0].render(script, [madfileset[0]] + x_conf)

        invoke_cmd = self.conf['invocator.execute_command']
        invoke_cmd = madfileset[0].render(invoke_cmd, [madfileset[0]] + x_conf)

        if ('{{' in cl) or ('{%' in cl):
            lg.error("cannot render command line")
            lg.error(" cl: %s", cl)
            exit(-1)

        if '{{' in script or '{%' in script:
            lg.error("cannot render execution script")
            lg.error(" cl: %s", script)
            exit(-1)

        if '{{' in invoke_cmd or '{%' in invoke_cmd:
            lg.error("cannot render invocation script")
            lg.error(" cl: %s", invoke_cmd)
            exit(-1)

        lg.debug("cl: {0}".format(cl))
        lg.debug(
            "script: {0}".format(script))
        lg.debug(
            "invocate: {0}".format(invoke_cmd))

        return cl, script, invoke_cmd

    def execute(self, madfileset, command):
        """
        Execute a single command line on a (set of) file(s)
        """

        def _applicator(invoke, script, bg):
            P = sp.Popen(invoke, shell=True, stdin=sp.PIPE)
            P.communicate(script)

        cl, script, invoke = \
            self.prepare_script(madfileset, command)

        if self.args.dry:
            print('#' * 80, 'invoke')
            print(invoke)
            print('#' * 80, 'execute')
            print(script)
            print('#' * 80)
        else:
            rv = fantail.Fantail()
            rv['cl'] = cl
            self.pool.apply_async(_applicator,
                                  (invoke, script, self.args.bg))
            return rv

    def finish(self):
        """
        Close & join the executor pool
        """
        self.pool.close()
        self.pool.join()
        lg.debug("finished executing")


def _get_command(app, madfile, command_name):

    filetype = madfile.get('filetype', "")


    if filetype:
        if command_name in app.conf['x.filetype'][filetype]:
            return app.conf['x.filetype'][filetype][command_name]

    #print(app.conf['x.filetype.default'].keys())
    return app.conf['x.filetype.default'][command_name]


@leip.arg('file', nargs='*')
@leip.command
def commands(app, args):
    """
    Commands available:
    """
    for madfile in get_all_mad_files(app, args):
        filetype = madfile.get('filetype', "")
        if filetype:
            print('{} ({})'.format(madfile['inputfile'], filetype))
        else:
            print('{} (unknown file format)'.format(madfile['inputfile']))

        for command_name in app.conf['x.filetype.default']:
            cinf = app.conf['x.filetype.default'][command_name]
            print(' g {0}: {1}'.format(command_name, cinf['description']))

        if not filetype:
            return

        for command_name in app.conf['x.filetype'][filetype]:
            cinf = app.conf['x.filetype'][filetype][command_name]
            print(' s {0}: {1}'.format(command_name,
                                       cinf.get('description', 'no description')))


@leip.arg('file', nargs='*')
@leip.arg('comm', metavar='command', help='predefined command to execute')
@leip.arg('-j', '--threads', help='no of threads', type=int, default=1)
@leip.arg('-x', '--executor', help="executor to use", default='simple',
          choices=['simple', 'pbs'])
@leip.arg('-g', '--groupby',
          help='group on this field for combined execution')
@leip.flag('-d', '--dry', help='dry run')
@leip.flag('-b', '--bg', help='run in the background')
@leip.command
def x(app, args):
    """
    Execute a command
    """
    command_name = args.comm

    executor = Executor(app, args)

    if args.groupby:
        group_on = args.groupby
        lg.info("Grouping on %s", group_on)
        groups = collections.defaultdict(list)
        for f in get_all_mad_files(app, args):
            groups[f[group_on]].append(f)
        files = groups.values()
    else:
        files = [[x] for x in get_all_mad_files(app, args)]

    #print(list(f[0]['filename'] for f in files))

    for madfileset in files:

        for madfile in madfileset:
            command_info = _get_command(app, madfile, command_name)
            if not command_info:
                lg.error("Command {} does not exists for {}".format(
                         command_name, madfile['filename']))
                exit(-1)

        runinfo = executor.execute(madfileset, command_info)

        # t = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
        # h = madfile.mad.execute.history[t]
        # h.update(runinfo)
        # madfile.save()

    executor.finish()



@leip.arg('-g', '--group', help='group on', default=1)
@leip.arg('file', nargs='*')
@leip.arg('comm', metavar='command', help='predefined command to execute')
@leip.command
def x2(app, args):
    lg.warning("x")

    groupby = args.group
    try:
        groupby = int(groupby)
        numeric_groups = True
    except:
        raise NotImplemented("Non numeric groups")

    mf_generator = get_all_mad_files(app, args)

    for res in render_numeric(app, mf_generator, args.comm, args.group):
        print(res)

# h()
# madfile.save()

#     executor.finish()
# h()




import datetime
import logging

from pymongo import MongoClient

lg = logging.getLogger(__name__)
#lg.setLevel(logging.DEBUG)

MONGO_SAVE_CACHE = []
MONGO_SAVE_COUNT = 0
MNG = None

#These fields we do not want to see in the core database
FORBIDDEN = ['hash', 'uuid', '_id_dump', 'host', 'volume']

def mongo_prep_mad(mf):

    d = dict(mf)

    mongo_id = mf['sha1sum'][:24]
    d['_id'] = mongo_id
    d['sha1sum'] = mf['sha1sum']
    d['save_time'] = datetime.datetime.utcnow()

    for f in FORBIDDEN:
        if f in d:
            del d[f]
    return mongo_id, d


class MongoStore():

    def __init__(self, conf):
        lg.debug("starting mongostore")
        self.conf = conf

        self.host = self.conf.get('host', 'localhost')
        self.port = int(self.conf.get('port', 27017))
        self.client = MongoClient(self.host, self.port)
        self.db_name = self.conf.get('db', 'mad2')
        self.corename = self.conf.get('collection', 'core')

        self.dumpname = self.conf.get('dump_collection', 'dump')

        self.db_core = self.client[self.db_name][self.corename]
        self.db_dump = self.client[self.db_name][self.dumpname]

        self.save_cache = []

    def prepare(self, madfile):
        return

    def changed(self, madfile):
        pass

    def delete(self, madfile):
        """
        We do not really delete a file here - but - this needs to be
        calleable
        """
        lg.debug("madfile delete %s", madfile['inputfile'])

    def save(self, madfile):
        """Save data to the mongo database"""

        if not 'sha1sum' in madfile:
            lg.warning("cannot save to mongodb without a sha1sum")
            return

        if madfile['sha1sum'] is None:
            lg.warning("cannot save to mongodb without a sha1sum")
            return

        if madfile['orphan']:
            lg.warning("Will not save non-existing files")
            return

        mongo_id = madfile['sha1sum'][:24]

        core = dict(madfile.mad)
        core['sha1sum'] = madfile['sha1sum']

        if 'hash' in core:
            del core['hash']
        if 'uuid' in core:
            del core['uuid']
        if '_id_dump' in core:
            del core['_id_dump']

        core['_id'] = mongo_id
        del core['_id']

        lg.debug("saving to id %s", mongo_id)
        self.save_cache.append((mongo_id, core))

        if len(self.save_cache) > 20:
            self.flush()

    def flush(self):

        if len(self.save_cache) == 0:
            return

        bulk = self.db_core.initialize_unordered_bulk_op()
        for i, r in self.save_cache:
            bulk.find({'_id': i}).upsert().replace_one(r)
        res = bulk.execute()
        lg.debug("Modified %d records", res['nModified'])
        self.save_cache = []


    def load(self, madfile, sha1sum):
        """
        Load the file from the databse,
        possibly with an alternative sha1sum
        """

        if not sha1sum is None:
            lg.warning("load from record with alternative cs %s", sha1sum)
            sha1 = sha1sum
        else:

            sha1 = madfile.get('sha1sum')

        if sha1 is None:
            return

        mongo_id = sha1[:24]
        lg.debug("getting mad data for {}".format(
                 madfile['inputfile']))
        lg.debug(" - sha1: {}".format(sha1))
        lg.debug(" - mongo_id: {}".format(mongo_id))
        data = self.db_core.find_one({'_id': mongo_id})
        madfile.mad.update(data)

    def finish(self):
        lg.debug("cleaning up")
        self.flush()




import copy
import os

from jinja2 import TemplateNotFound
from jinja2 import Environment, PackageLoader

import leip
from mad2.plugin import mongo
from mad2.ui import message

@leip.subparser
def report(app, args):
    """
    Report generation
    """
    pass


@leip.flag('-f', '--force')
@leip.arg('dir', nargs='?', default='.', help='location to generate the '
          'static html files' )
@leip.subcommand(report, "generate")
def report_generate(app, args):
    """
    Generate html reports
    """
    direc = os.path.expanduser(args.dir)

    env = Environment(
            loader=PackageLoader('mad2', 'template/report'))

    #make directory structure
    def mkdir(*args):
        p = os.path.join(*args)
        if not os.path.exists(p):
            os.makedirs(p)

    mkdir(direc, 'data')

    conf = app.conf['plugin.report']
    singletons = conf['singletons']

    data = {'singletons': singletons,
             'singleton': ''
            }

    def rendersave(name, tname=None, data={}, **kwargs):
        d = copy.copy(data)
        d.update(kwargs)

        message("creating %s" % name)
        if tname is None:
            template = env.get_template(name)
        else:
            template = env.get_template(tname)
        with open(os.path.join(args.dir, name), 'w') as F:
            F.write(template.render(d))

    rendersave('mad.css')
    print(data)
    rendersave('index.html', data=data)

    #generate singleton data
    for s in data['singletons']:
        message("getting data for %s", s)
        sumdata = mongo._single_sum(app, group_by=s, force=args.force)
        tsv_file = os.path.join(direc, 'data', 'single_{}.tsv'.format(s))
        with open(tsv_file, 'w') as F:
            F.write('name\tcount\tsum\n')
            for rec in sumdata:
                F.write('{_id}\t{count}\t{total}\n'.format(**rec))
            rendersave('single_{}.html'.format(s),
                       tname='singleton.html', data=data, singleton=s)







from __future__ import print_function, division

from datetime import datetime
import os
import logging
import re
import socket

import leip
import fantail
from pwd import getpwuid

lg = logging.getLogger(__name__)

EXTENSION_DATA = None
RECURSE_CACHE = {}


def get_fiex(app):
    global EXTENSION_DATA
    if not EXTENSION_DATA is None:
        return EXTENSION_DATA

    EXTENSION_DATA = {}

    for ft_name in app.conf['filetype']:
        ft = app.conf['filetype'][ft_name]
        for ext in ft.get('extensions', []):
            EXTENSION_DATA[ext] = ft_name, ft

    return EXTENSION_DATA


def apply_file_format(app, madfile, filename=None):

    extension_data = get_fiex(app)
    if filename is None:
        filename = madfile['filename']

    splitter = filename.rsplit('.', 1)
    if len(splitter) != 2:
        return

    base, ext = splitter

    madfile.all['basename'] = base
    # this ensures that the innermost extension seen is stored
    madfile.all['extension'] = ext

    if not ext in extension_data:
        return

    filetype, ftinfo = extension_data[ext]
    lg.debug("identified filetype {0}".format(filetype))
    overlay_name = ftinfo.get('overlay')

    overlay = app.conf['overlay.{0}'.format(overlay_name)]
    madfile.all.update(overlay)
    if ftinfo.get('continue', False):
        lg.debug("contiue filetype disocvery on: %s" % base)
        apply_file_format(app, madfile, base)


def _get_recursive_dir_data(pth):

    global RECURSE_CACHE
    lg.debug("start recursive data load for {}".format(pth))

    pth = os.path.abspath(pth)

    if os.path.exists(pth) and not os.path.isdir(pth):
        here = os.path.dirname(pth)
    else:
        here = pth

    here = here.rstrip('/')
    conf = []

    # find existing directory configuration from the perspective
    # of the madfile
    last = here

    while True:

        if os.path.isdir(here):

            here_c = os.path.join(here, 'mad.config')
            if os.path.exists(here_c):
                conf.append(here_c)

        parent = os.path.dirname(here)

        if parent == '/':  # no config in the root - that would be evil!
            break
        last = here
        here = parent

    # now again from the current directory
    here = os.getcwd().rstrip('/')
    last = here
    cwdconf = []
    while True:
        assert(os.path.isdir(here))

        here_c = os.path.join(here, 'mad.config')
        if os.path.exists(here_c):
            if here_c in conf:
                break  # overlap with tree from the madfile's location
            else:
                cwdconf.append(here_c)
        parent = os.path.dirname(here)
        if parent == '/':  # no config in the root - that would be evil!
            break
        last = here
        here = parent

    conf = cwdconf + conf
    # load (or get from cache)

    rv = fantail.Fantail()
    for c in conf[::-1]:
        fullname = os.path.expanduser(os.path.abspath(c))
        if fullname in RECURSE_CACHE:
            y = RECURSE_CACHE[fullname]
        else:
            #print('start load dir', fullname)
            y = fantail.yaml_file_loader(fullname)
            RECURSE_CACHE[fullname] = y
        rv.update(y)
    return rv

@leip.arg("dir", nargs='?', default='.')
@leip.command
def project(app, args):
    rdd = _get_recursive_dir_data(args.dir)
    project = rdd.get('project', '')
    if len(project) > 0:
        print(project)


@leip.hook("madfile_pre_load")
def recursive_dir_data(app, madfile):
    y = _get_recursive_dir_data(madfile['fullpath'])
    # insert in the stack just after the mad file
    madfile.all.update(y)

    keywords = app.conf['keywords']

    #add all non transient keys to the core database
    for k in y:
        if keywords[k]['transient']:
            continue
        madfile.mad[k] = y[k]


@leip.hook("madfile_pre_load")
def onthefly(app, madfile):

    lg.debug("running onthelfy")

    lg.debug("get fqdn")
    host = socket.gethostname()
    madfile.all['host'] = host
    lg.debug('host: %s', host)

    madfile.all['volume'] = host

    madfile.all['uri'] = "file://{}{}".format(
        madfile.all['host'], madfile['fullpath'])

    if madfile.get('orphan', False):
        # orphaned is file - little we can do
        return

    #    if not os.path.exists(madfile['fullpath'])
    filestat = os.lstat(madfile['fullpath'])
    # print(filestat)

    madfile.all['filesize'] = filestat.st_size
    madfile.all['nlink'] = filestat.st_nlink
    madfile.all['is_symlink'] = os.path.islink(madfile['fullpath'])

    try:
        userinfo = getpwuid(filestat.st_uid)
    except KeyError:
        # cannot find username based on uid
        madfile.all['userid'] = str(filestat.st_uid)
        madfile.all['username'] = str(filestat.st_uid)
    else:
        madfile.all['userid'] = userinfo.pw_name
        madfile.all['username'] = userinfo.pw_gecos

    # if not app.conf.get('username') is None:
    #     madfile.all['username'] = app.conf['username']

    mtime = datetime.utcfromtimestamp(
                filestat.st_mtime)
    atime = datetime.utcfromtimestamp(
        filestat.st_atime)

    madfile.all['atime'] = atime
#    madfile.all['atime_simple'] = atime.strftime("%Y/%m/1")

    madfile.all['mtime'] = mtime
#    madfile.all['mtime_simple'] = mtime.strftime("%Y/%m/1")
    madfile.all['basename'] = madfile.all['filename']

    apply_file_format(app, madfile)

    thesaurus = app.conf['thesaurus']
    for t in thesaurus.values():
        if len(t['find']) != 1:
            lg.critical("cannot handle multiple search fields")
            exit(0)
        f_field = t['find'].keys()[0]
        f_pattern = t['find'].values()[0]
        replace = t['replace']
        if f_field in madfile and re.match(f_pattern, madfile[f_field]):
            #match - now update the
            madfile.all.update(replace)

    # last - but not least
    # make sure the file has a SHA1SUM, and that it is up to date


    madfile.check_sha1sum()

    lg.debug("finished onthefly")

@leip.arg('alias')
@leip.arg('userid')
@leip.command
def user_alias(app, args):
    lg.info("setting user alias: '%s' to '%s'", args.userid, args.alias)
    loco = leip.get_local_config_file('mad2')
    thes = loco['thesaurus']

    #always use the same id!
    lid = 'user_' + args.userid

    thes[lid]['find.userid'] = args.userid
    thes[lid]['replace.username'] = args.alias
    #print(loco.pretty())

    leip.save_local_config_file(loco, 'mad2')
    leip.get_config('mad2', rehash=True)


@leip.arg('volume_name')
@leip.arg('path_fragment')
@leip.command
def volume_alias(app, args):

    loco = leip.get_local_config_file('mad2')
    thes = loco['thesaurus']

    #always use the same id!
    lid = 'path_' + re.sub('\W+', '_', args.path_fragment).strip("_")
    rex = '.*' + re.escape(args.path_fragment) + '.*'
    thes[lid]['find.fullpath'] = rex
    thes[lid]['replace.volume'] = args.volume_name
    print(loco.pretty())

    leip.save_local_config_file(loco, 'mad2')
    lgf = logging.getLogger("fantail.util")
    lgf.setLevel(logging.INFO)
    leip.get_config('mad2', rehash=True)



#!/usr/bin/env python


import logging
import os
import sys
import socket


logging.basicConfig()

lg = logging.getLogger(__name__)
lg.setLevel(logging.DEBUG)


if len(sys.argv) > 1:
    start = sys.argv[1]
else:
    start = '.'

start = os.path.fullpath(start)

lg.info("start dir: {}".format(start))


base = dict(
    socket=socket.gethostname()
)


def process_dir(dir):
    lg.info("processing: {}".format(dir))

    #if (dir / 'mad.ignore')
    for file in dir.files():
        if file.basename()[0] == '.':
            continue
        yield file


    for sub in dir.dirs():
        if sub.basename()[0] == '.':
            continue
        for file in process_dir(sub):
            yield file

def read_cache
for dirpath, dirnames, filenames in os.walk(start):
    print(dirpath)





    #
    #
    # current = Path(currentpath)
    #
    # to_remove = []
    # for d in dirs:
    #     if d[0] == '.':
    #         to_remove.append(d)
    #     if os.path.exists(os.path.join(currentpath, d, 'mad.ignore')):
    #         to_remove.append(d)
    #
    # for d in to_remove:
    #     dirs.remove(d)
    #
    # for f in files:
    #     if f[0] == '.':
    #         continue
    #
    #     filesize =
    #
    #
    # if dirs:
    #     print(dirs)







import logging
import os

import fantail

from mad2.exception import MadPermissionDenied

lg = logging.getLogger(__name__)
#lg.setLevel(logging.DEBUG)


class SidecarStore():

    def __init__(self, conf):
        self.conf = conf

    def changed(self, madfile):
        pass

    def delete(self, madfile):
        lg.warning("should delete: %s", madfile['inputfile'])

    def flush(self):
        pass

    def prepare(self, madfile):

        inputfile = madfile['inputfile']
        filename = madfile['filename']
        dirname = madfile['dirname']

        if madfile.dirmode:
            lg.warning('no dirs')
            return #not doing directories anymore

        # looking at file
        if filename[-4:] == '.mad':

            if filename[0] == '.':
                filename = filename[1:-4]

            madname = inputfile
            inputfile = os.path.join(dirname, filename)
        else:
            inputfile = inputfile
            madname = os.path.join(dirname, '.' + filename + '.mad')

        lg.debug("madname: {}".format(madname))
        lg.debug("inputfile: {}".format(inputfile))

        madfile.all['madname'] = madname
        madfile.all['fullmadpath'] = os.path.abspath(madname)

        if os.path.exists(madname) and not os.access(madname, os.R_OK):
            raise MadPermissionDenied()

        if madfile.get('orphan', False) and os.path.exists(madname):
            lg.warning("Orphaned mad file: {}".format(madname))
            lg.debug("  | can't find: {}".format(inputfile))

    def save(self, madfile):

        if self.conf.get('readonly'):
            return

        # if self.
        try:
            lg.debug("saving to %s" % madfile['madname'])
            # note the mad file data is in stack[1] - 0 is transient
            # print(self.mad)
            fantail.yaml_file_save(madfile.mad, madfile['madname'])
        except IOError, e:
            if e.errno == 36:
                lg.error("Can't save - filename too long: {}"
                         .format(self.fullpath))
            else:
                raise

    def load(self, madfile, sha1sum=None):
        """

        sha1sum is never used for sidecars

        """

        lg.debug("sidecar load: %s", madfile)

        if os.path.exists(madfile['madname']):
            lg.debug("loading madfile {0}".format(madfile['madname']))

            # note the mad file data is in stack[1] - 0 is transient
            madfile.mad.update(
                fantail.yaml_file_loader(madfile['madname']))


    def finish(self):
        lg.debug("cleaning up")





from __future__ import print_function

import logging
import leip
import os
import re

lg = logging.getLogger(__name__)

@leip.hook("madfile_post_load", 100)
def lbconthefly(app, madfile):

    fullpath = madfile['fullpath']
    mtcher = re.compile(r'/media/(seq-srv-[0-9][0-9])', re.I)
    mtch = mtcher.search(fullpath)
    if not mtch:
        return


    server = mtch.groups()[0]

    #TDDO: remove this later - bugfix
    if 'host' in madfile.mad:
        del madfile.mad['host']
    madfile.all['host'] = server.upper()
